#!/usr/bin/env python
# encoding: utf-8

import numpy as np
from scipy.fftpack import fft,fft2,ifft2
import matplotlib.pyplot as plt
import imageio
import os
class PDESpectralSolver(object):
    def __init__(self, laplace, cc):
        """
        Solving 2,4,... order differential equation 
        $$ c0*u + c1*\Delta^1 u + c2*\Delta^2 u + ... = f $$ 
        $cc = [c0, c1, c2, ...]$
        periodic boundary by Fourier spectral method
        """
        if len(cc) == 2:
            self.derive = cc[0] + cc[1]*laplace
        elif len(cc) == 3:
            self.derive = cc[0] + cc[1]*laplace + cc[2]*laplace*laplace
        else:
            self.derive = np.zeros_like(laplace)
            for k, ck in enumerate(cc):
                self.derive += ck*laplace**k
            
    def getSolution(self, fmesh):
        f_f = fft2(fmesh)
        return ifft2(f_f/self.derive).real

    def getSolutionFF(self, ff):
        return ifft2(ff/self.derive).real

### mesh
Nx = 128
Ny = 128
x = np.linspace(-1, 1, Nx, False)
y = np.linspace(-1, 1, Ny, False)
mesh = np.meshgrid(x, y)

### laplace coefficent
kkx = np.hstack((range(Nx//2), [0], range(-Nx//2+1, 0)))
kky = np.hstack((range(Ny//2), [0], range(-Ny//2+1, 0)))
k2d = np.meshgrid(1j*np.pi*kkx, 1j*np.pi*kky);
laplace = k2d[0]**2 + k2d[1]**2

u = lambda x: np.sin(np.pi*(x[0]))*np.sin(np.pi*(x[1]))

"""
### test PDESpectralSolver 
cc1 = [2, -3]
cc2 = [4, -3, 2]
lapc = -2*np.pi**2
f1 = lambda x: (cc1[0] + cc1[1]*lapc)*u(x)
f2 = lambda x: (cc2[0] + cc2[1]*lapc + cc2[2]*lapc**2)*u(x)
### test 1 order
solver1 = PDESpectralSolver(laplace, cc1)
uN1 = solver1.getSolution(f1(mesh))
err1 = uN1 - u(mesh)
print(abs(err1).max())
### test 2 order
solver2 = PDESpectralSolver(laplace, cc2)
uN1 = solver2.getSolution(f2(mesh))
err2 = uN1 - u(mesh)
print(abs(err2).max())
"""

def uxt(t):
    return lambda x: np.sin(np.pi*(x[0]))*np.sin(np.pi*(x[1]))*np.exp(-t)


#### AC equation
def h(phi):
    return phi*(phi*phi - 1)
    # tmp = 2*np.pi**2
    # return (1-tmp*tmp)/tmp*phi
t0 = 0.0
MaxTStep = 1000
dt = 1.0/MaxTStep
T = dt*MaxTStep
stablecoef = 1
solver = PDESpectralSolver(laplace, [1, -dt*stablecoef, dt])
phin = uxt(t0)(mesh)

plt.subplots()
plt.contourf(mesh[0], mesh[1], phin)
kkk = 0
filename = "tmp.png"
images = []
for i in range(MaxTStep):
    lapphi = ifft2(laplace*fft2(phin)).real
    laphphi = ifft2(laplace*fft2(h(phin))).real
    tmp = phin - dt*stablecoef*lapphi + dt*laphphi
    phin = solver.getSolution(tmp)
    if i%10 == 0:
        kkk += 1
        plt.contourf(mesh[0], mesh[1], phin)
        plt.savefig(filename, dip=6)  # dip表示图片的质量，数值越大图片质量越高
        images.append(imageio.imread(filename))
        
imageio.mimsave('CHtest.gif', images, duration=0.6)
if os.path.exists(filename):
    os.remove(filename)
    
# err = phin - uxt(T)(mesh)
# print(abs(err).max())



#### AC equation
# def h(phi):
    # tmp = 2*np.pi**2
    # return (1-tmp)*phi
# t0 = 0.0
# MaxTStep = 10
# dt = 1.0/MaxTStep
# T = dt*MaxTStep
# stablecoef = 1
# solver = PDESpectralSolver(laplace, [1+dt*stablecoef, -dt])
# phin = uxt(t0)(mesh)
# for i in range(MaxTStep):
    # phin = solver.getSolution(phin*(1+dt*stablecoef)-dt*h(phin))
# err = phin - uxt(T)(mesh)
# print(abs(err).max())





### heat equation
# t0 = 0.0
# MaxTStep = 10000
# dt = 1.0/MaxTStep
# T = dt*MaxTStep
# stablecoef = 10
# solver = PDESpectralSolver(laplace, [1+dt*stablecoef, -dt/(2*np.pi**2)])
# phin = uxt(t0)(mesh)
# for i in range(MaxTStep):
    # phin = solver.getSolution(phin*(1+dt*stablecoef))
# err = phin - uxt(T)(mesh)
# print(abs(err).max())




"""
def uxt(t):
    return lambda x: np.sin(np.pi*(x[0]))*np.sin(np.pi*(x[1]))*np.exp(-t)
def H(phi):
    return 0.25*(phi*phi - 1)*(phi*phi - 1)
    
# def h(phi):
    # return (phi*phi - 1)*phi

def h(phi):
    tmp = 2*np.pi**2
    return (1-tmp*tmp)/tmp*phi
    
fint = lambda fu : fft2(fu)[0][0].real*(4.0/(Nx*Ny))
## see slides http://kz16.top/nonlinear/ch/#/9
t0 = 0.0
dt = 0.01
MaxTStep = 1
T = dt*MaxTStep
stablecoef = 0
solver = PDESpectralSolver(laplace, dt, 1, s, stablecoef*dt)
phin = u(mesh)
xi = 1



plt.subplots()
plt.contourf(mesh[0], mesh[1], phin)
phis = np.zeros((MaxTStep//50, Nx, Ny))

kkk = 0
for i in range(MaxTStep):
    phi1 = solver.getSolution(phin)
    tmp = ifft2(laplace*fft2(dt*h(phin) - dt*stablecoef*phin))
    phi2 = solver.getSolution((tmp).real)
    phinp1 = phi1 + xi*phi2

    
    phin = phinp1
    
    # if i%100 == 0 or i < 20:
        # phis[kkk] = phin
        # kkk += 1
        # plt.cla()
        # plt.contourf(mesh[0], mesh[1], phin)
        # plt.show()
        
err = phin - uxt(T)(mesh)
print(abs(err).max())


### quad test

fint = lambda fu : fft2(fu)[0][0].real*(4.0/(Nx*Ny))
testu = lambda x: x[0]*x[1]*x[0]*x[1]
print(fint(testu(mesh)))
u = testu(mesh)
print(u.sum()*(4.0/(Nx*Ny)))
print(4.0/9)
"""






